from __future__ import annotations
from dataclasses import dataclass, field
import typing
import importlib

import datetime as dt
from patent_client.util import Model, one_to_one, one_to_many, QuerySet
from .lookups import lookup_claims, lookup_description, lookup_family

# Core INPADOC Models

@dataclass
class Inpadoc(Model):
    __manager__ = 'patent_client.epo.inpadoc.manager.InpadocManager'
    number: str
    doc_type: str
    kind_code: typing.Optional[str] = None
    country: typing.Optional[str] = None
    family_id: typing.Optional[str] = None
    date: typing.Optional[dt.date] = None

    biblio = one_to_one('patent_client.epo.inpadoc.model.InpadocBiblio', publication='num')
    claims = lookup_claims()
    description = lookup_description()
    family = lookup_family()

    @property
    def num(self):
        return f"{self.country}{self.number}{self.kind_code}"
    
    @property
    def us_application(self):
        klass = getattr(importlib.import_module('patent_client.uspto.peds.model'), 'USApplication')
        if self.kind_code == 'A1':
            pub_num = self.country + self.number[:4] + self.number[4:].rjust(7, '0') + self.kind_code
            return klass.objects.get(app_early_pub_number=pub_num)
        else:
            return klass.objects.get(patent_number=self.number)
        
@dataclass
class InpadocPublication(Inpadoc):
    biblio = one_to_one('patent_client.epo.inpadoc.model.InpadocBiblio', publication='num')

@dataclass
class InpadocApplication(Inpadoc):
    biblio = one_to_one('patent_client.epo.inpadoc.model.InpadocBiblio', application='num') 

# INPADOC Detail Models

@dataclass
class InpadocPriorityClaim(Inpadoc):
    children = one_to_many('patent_client.epo.inpadoc.model.Inpadoc', priority_claim='num')
    biblio = one_to_one('patent_client.epo.inpadoc.model.InpadocBiblio', application='num')  

@dataclass
class InpadocBiblio(Model):
    __manager__ = 'patent_client.epo.inpadoc.manager.InpadocBiblioManager'
    family_id: str
    title: str
    number: str
    kind_code: str
    country: str
    applications: typing.List['InpadocApplication']
    publications: typing.List['InpadocPublication']
    applicants: typing.List[str]
    inventors: typing.List[str]
    abstract: str = None
    ipc_classes: typing.List[str] = field(default_factory=list)
    cpc_classes: typing.List[CpcClass] = field(default_factory=list)
    us_classes: typing.List[str] = field(default_factory=list)
    priority_claims: QuerySet[InpadocPriorityClaim] = field(default_factory=list)

    @property
    def num(self):
        return f"{self.country}{self.number}{self.kind_code}"


# INPADOC Family Models

@dataclass
class InpadocFamilyPriorityClaim(Model):
    doc: Inpadoc
    active: bool
    seq: int
    kind: str
    active: str
    link_type: typing.Optional[str] = None

@dataclass
class InpadocFamilyMember(Model):
    publication: 'InpadocPublication'
    application: 'InpadocApplication'
    priority_claims: typing.List['InpadocPriorityClaim']
    family_id:str 

# Classification Models

@dataclass
class CpcClass(Model):
    sequence: str
    section: str
    cpc_class: str
    subclass: str
    main_group: str
    sub_group: str
    classification_value: str
    generating_office: str








import importlib
import logging

logger = logging.getLogger(__name__)

from .manager import QuerySet, resolve


def one_to_one(class_name, attribute=None, **mapping):
    module_name, class_name = class_name.rsplit(".", 1)

    @property
    def get(self):
        try:
            klass = getattr(importlib.import_module(module_name), class_name)
            filter_obj = {k: getattr(self, v) for (k, v) in mapping.items()}
            logger.debug(f'Fetching related {klass} using filter {filter_obj}')
            return resolve(klass.objects.get(**filter_obj), attribute)
        except AttributeError:
            return None

    return get


def one_to_many(class_name, **mapping):
    module_name, class_name = class_name.rsplit(".", 1)

    @property
    def get(self):
        klass = getattr(importlib.import_module(module_name), class_name)
        filter_obj = {k: getattr(self, v) for (k, v) in mapping.items()}
        return klass.objects.filter(**filter_obj)

    return get

def get_manager(class_name):
    module_name, class_name = class_name.rsplit(".", 1)
    @property
    def objects(self):
        klass = getattr(importlib.import_module(module_name), class_name)
        return klass()
    return objects



def recur_accessor(obj, accessor):
    if "__" not in accessor:
        a = accessor
        rest = None
    else:
        a, rest = accessor.split("__", 1)

    if hasattr(obj, a):
        new_obj = getattr(obj, a)
        if callable(new_obj):
            new_obj = new_obj()
    else:
        try:
            a = int(a)
        except ValueError:
            pass
        try:
            new_obj = obj[a]
        except (KeyError, TypeError, IndexError):
            new_obj = None
    if not rest:
        return new_obj
    else:
        return recur_accessor(new_obj, rest)



import pytest
from .model import PtabProceeding, PtabDocument, PtabDecision

class TestPtabProceeding:
    def test_get_by_proceeding_number(self):
        result = PtabProceeding.objects.get("IPR2016-00831")
        assert result.patent_number == "6162705"
    
    def test_get_by_patent_number(self):
        result = PtabProceeding.objects.get(patent_number='6103599')
        assert result.proceeding_number == "IPR2016-00833"

    def test_get_by_application_number(self):
        result = PtabProceeding.objects.get(appl_id='09089931')
        assert result.proceeding_number == "IPR2016-00833"

    def test_filter_by_party(self):
        result = PtabProceeding.objects.filter(party_name="Apple")
        assert len(result) >= 400
        assert result.count() >= 400

    def test_filter_with_limit(self):
        result = PtabProceeding.objects.filter(party_name="Apple").limit(26)
        assert len(result) == 26
        objects = list(result)
        assert len(objects) == 26

    def test_offset(self):
        result = PtabProceeding.objects.filter(party_name="Apple").limit(3)
        objects = list(result)
        assert result.first() == objects[0]
        assert result.offset(1).first() == objects[1]
        assert result.offset(1).offset(1).first() == objects[2]

class TestPtabDocument:
    def test_filter_by_proceeding(self):
        result = PtabDocument.objects.filter(proceeding_number="IPR2016-00831")
        assert len(result) == 78

    def test_sort_by_document_number(self):
        result = (
            PtabDocument.objects.filter(proceeding_number="IPR2016-00831")
            .order_by("document_number")
            .limit(3)
            .values_list("document_number", flat=True)
        )
        assert len(result) == 3
        objects = list(result)
        assert objects == list(sorted(objects))

class TestPtabDecision:
    def test_get_by_proceeding(self):
        result = PtabDecision.objects.get(proceeding_number="IPR2016-00831")
        assert (
            result.identifier
            == "a44c5f1557b7b60d00e66604d3668ce442d53f964aa597011cc476b4"
        )


from __future__ import annotations
import itertools
import importlib
from copy import deepcopy
from collections import OrderedDict
import collections.abc
from typing import Iterable, Iterator, TypeVar, Generic, Union, Sized, List, Any

ModelType = TypeVar('ModelType')

def resolve(item, key):
    if key is None:
        return item
    accessors = key.split(".")
    try:
        for accessor in accessors:
            if accessor.isdigit():
                item = item[int(accessor)]
            elif isinstance(item, dict):
                item = item[accessor]
            else:
                item = getattr(item, accessor)
            if callable(item):
                item = item()
    except Exception as e:
        return None 
    return item

class Manager(Generic[ModelType]):
    """
    Manager Base Class

    This class is essentially a configurable generator. It is intended to be initialized 
    as an empty object at Model.objects. Users can then modify the manager by calling either:
        
        Manager.filter -> adds filter criteria
        Manager.sort -> adds sorting criteria
        Manager.options -> adds key:value options

    All methods should return a brand-new manager with the appropriate parameters re-set. The 
    Manager also has a special method to fetch a single matching object:

        Manager.get -> adds filter critera, and returns the first object if only one object is found, else raises an Exception

    The manager's attributes are stored in a dictionary at Manager.config. Config has the following structure:

        {
            'filter': {
                key: value,
                key: value,
            },
            'order_by': [value, value, value],
            'values': [value, value, value]
            'options': {
                key: value
            }
        }
        
    """
    primary_key: str = ''

    def __init__(
        self,
        config=dict(
            filter=dict(), order_by=list(), options=dict(), limit=None, offset=0, annotations=list()
        ),
    ):
        self.config = config


    def __iter__(self) -> Iterator[ModelType]:
        """This function implements application-level caching
        The cache is held in the __cache__ attribute,
        The un-finished iterator/generator is held in the __result_iterator__ attribute
        The method it expects to call to get the iterator is self._get_results()
        """
        if self.config.get('disable_cache', False):
            # Caching can be disabled via option
            for item in self._get_results():
                yield item
        else:
            if not hasattr(self, '__result_iterator__'):
                # Create a new iterator and cache if the iterator doesn't exist
                self.__result_iterator__ = self._get_results()
                self.__cache__: List[ModelType] = list()
            # Yield out of the cache
            for item in self.__cache__:
                yield item
            # Yield out of the iterator, expanding the cache as you go.
            for item in self.__result_iterator__:
                self.__cache__.append(item)
                yield item

    def _get_results(self) -> Iterator[ModelType]:
        raise NotImplementedError('Must be implemented by subclass')

    def __len__(self) -> int:
        # The default len function runs the iterator and counts. There may be 
        # more efficient ways to do it for any given subclass, but this is the
        # basic way
        return len(list(self))

    def __add__(self, other: Manager) -> QuerySet:
        return QuerySet([self, other])

    def __eq__(self, other) -> bool:
        return bool(self.config == other.config and type(self) == type(other))

    def __getitem__(self, key:Union[slice, int]) -> Union[Manager[ModelType], ModelType]:
        if isinstance(key, slice):
            if key.step != None:
                raise AttributeError("Step is not supported")
            start = key.start if key.start else 0
            start = len(self) + start if start < 0 else start
            stop = key.stop if key.stop else len(self)
            stop = len(self) + stop if stop < 0 else stop
            mger = self.offset(start + self.config["offset"])
            mger = mger.limit(stop - start)
            return mger
        return self.offset(key).first()

    def filter(self, *args, **kwargs) -> Manager[ModelType]:
        if args:
            kwargs[self.primary_key] = args
        for k, v in kwargs.items():
            if not isinstance(v, str):
                try:
                    kwargs[k] = list(str(i) for i in v)
                except TypeError:
                    kwargs[k] = str(v)
            else:
                kwargs[k] = [v]
        new_config = deepcopy(self.config)
        new_config["filter"] = {**new_config["filter"], **kwargs}
        return self.__class__(new_config)

    def order_by(self, *args) -> Manager[ModelType]:
        """Take arguments, and store in a special keyword argument called 'sort' """
        new_config = deepcopy(self.config)
        new_config["order_by"] = list(new_config["order_by"]) + list(args)
        return self.__class__(new_config)

    def set_options(self, **kwargs) -> Manager[ModelType]:
        new_config = deepcopy(self.config)
        new_config["options"] = {**new_config["options"], **kwargs}
        return self.__class__(new_config)

    def _clone(self):
        return self.__class__(deepcopy(self.config))

    def limit(self, limit) -> Manager[ModelType]:
        clone = self._clone()
        clone.config['limit'] = limit
        return clone

    def offset(self, offset) -> Manager[ModelType]:
        clone = self._clone()
        clone.config['offset'] = self.config['offset'] + offset
        return clone

    def get(self, *args, **kwargs) -> ModelType: 
        """Implement a new manager with the requested keywords, and if the length is 1,
        return that record, else raise an exception"""
        manager = self.filter(*args, **kwargs)
        if len(manager) > 1:
            raise ValueError("More than one document found!")
        return manager[0] # type: ignore

    # Manager Functions

    def count(self) -> int:
        return len(self)

    def first(self) -> ModelType:
        return next(iter(self))

    def all(self) -> Manager[ModelType]:
        return self

    def to_pandas(self, annotate=list()):
        import pandas as pd
        dicts = list()
        for i in iter(self):
            ordered_dict = OrderedDict(i)
            for a in annotate:
                ordered_dict[a] = resolve(i, a)
            dicts.append(ordered_dict)
        return pd.DataFrame.from_records(dicts)

    def explode(self, attribute) -> QuerySet:
        from itertools import chain
        return ListManager(chain.from_iterable(getattr(r, attribute, None) for r in self))

    def to_records(self) -> Iterator[OrderedDict]:
        for i in self:
            yield i.as_dict()
    
    def to_list(self) -> List[ModelType]:
        return list(self)
    
    # Values
    def values(self, *fields, **kw_fields) -> ValuesQuerySet:
        return ValuesQuerySet(self, *fields, **kw_fields)

    def values_list(self, *fields, flat=False, **kw_fields) -> ValuesListQuerySet:
        return ValuesListQuerySet(self, *fields, flat=flat, **kw_fields)

class ListManager(list, Manager[ModelType]):
    
    def _clone(self):
        return self

    def limit(self, limit):
        return ListManager(self[:limit])

    def offset(self, offset):
        return ListManager(self[offset:])

    def __getitem__(self, sl):
        result = list(self)[sl]
        if isinstance(sl, slice):
            return ListManager(result)
        else:
            return result

    @classmethod
    def empty(cls) -> ListManager:
        return cls(list())

class QuerySet(Manager[ModelType]):
    """
    Utility class that extends the Manager helper function to 
    any collection of Patent Client objects
    """

    def __init__(self, managers, limit=None, offset=0):
        if isinstance(managers, Manager):
            self.managers = [managers._clone(),]
        else:
            self.managers = [m._clone() for m in managers]
        self._limit = limit
        self._offset = offset

    def __getitem__(self, key:Union[slice, int]) -> Union[Manager[ModelType], ModelType]:
        if isinstance(key, slice):
            if key.step != None:
                raise AttributeError("Step is not supported")
            start = key.start if key.start else 0
            start = len(self) + start if start < 0 else start
            stop = key.stop if key.stop else len(self)
            stop = len(self) + stop if stop < 0 else stop
            mger = self.offset(start + self._offset)
            mger = mger.limit(stop - start)
            return mger
        return self.offset(key).first()

    def __iter__(self):
        if self.offset == 0 and self.limit == None:
            for manager in self.managers:
                yield from iter(manager)
        else:
            counter = 0
            max_items = self._offset + self._limit if self._limit else None
            for manager in self.managers:
                if counter + len(manager) < self._offset or (self._limit and counter >= self._limit):
                    """In these circumstances, don't yield objects"""
                    counter += len(manager)
                    continue
                else:
                    offset = max(self._offset - counter, 0)
                    manager = manager.offset(offset)
                    if self._limit:
                        limit = self._limit - counter
                        print(limit)
                        manager = manager.limit(limit)
                    yield from manager
                    counter += len(manager)

    def __len__(self):
        return sum(len(i) for i in self.managers)

    def __repr__(self):
        return f"<QuerySet({self.managers})>"

    def _clone(self):
        return self.__class__(self.managers)

    def limit(self, limit):
        clone = self._clone()
        clone._limit = limit
        return clone
    
    def offset(self, offset):
        clone = self._clone()
        clone._offset = clone._offset + offset
        return clone

    @classmethod
    def empty(cls) -> QuerySet:
        return cls(list())

class ValuesQuerySet(QuerySet):
    def __init__(self, managers, *fields, **kw_fields):
        super(ValuesQuerySet, self).__init__(managers)
        self.fields = {
            **{k:k for k in fields},
            **kw_fields
        }
    
    def _clone(self):
        return self.__class__(self.managers, **self.fields)

    def __iter__(self) -> Iterator[OrderedDict]:
        for item in super(ValuesQuerySet, self).__iter__():
            yield OrderedDict((k, resolve(item, v)) for k, v in self.fields.items())

class ValuesListQuerySet(ValuesQuerySet):
    def __init__(self, managers, *fields, flat=False, **kw_fields):
        super(ValuesListQuerySet, self).__init__(managers, *fields, **kw_fields)
        self.flat = flat
    
    def _clone(self):
        return self.__class__(self.managers, flat=self.flat, **deepcopy(self.fields))

    def __iter__(self):
        if self.flat and len(self.fields) > 1:
            raise ValueError("Flat only works with 1 field!")
        for data_dict in super(ValuesListQuerySet, self).__iter__():
            data = tuple(data_dict.values())
            yield data[0] if self.flat else data



from marshmallow import Schema, fields, EXCLUDE, pre_load, post_load
from patent_client.util.manager import resolve
from patent_client.util import QuerySet, ListField

from .model import Inpadoc, CpcClass, InpadocBiblio, InpadocApplication, InpadocPublication, InpadocPriorityClaim

class BaseSchema(Schema):
    @post_load
    def make_object(self, data, **kwargs):
        if hasattr(self, '__model__'):
            return self.__model__(**data)
        return data

class InpadocResultSchema(BaseSchema):
    __model__ = Inpadoc
    country = fields.Str(data_key='country')
    number = fields.Str(data_key='doc-number')
    kind_code = fields.Str(data_key='kind')
    doc_type = fields.Str(data_key='@document-id-type')
    family_id = fields.Int(data_key='@family-id')

    @pre_load
    def pre_load(self, data, *args, **kwargs):
        for k, v in data['document-id'].items():
            data[k] = v
        del data['document-id']
        return data

    
    class Meta:
        unknown=EXCLUDE

def flatten(data, subfield):
    data = {**data, **data[subfield]}
    del data[subfield]
    return data

class InpadocSchema(BaseSchema):
    __model__ = Inpadoc
    country = fields.Str(data_key='country')
    number = fields.Str(data_key='doc-number')
    kind_code = fields.Str(data_key='kind')
    doc_type = fields.Str(data_key='@document-id-type')
    date = fields.Date(format='%Y%m%d')

class InpadocApplicationSchema(InpadocSchema):
    __model__ = InpadocApplication
    class Meta:
        unknown=EXCLUDE

class InpadocPublicationSchema(InpadocSchema):
    __model__ = InpadocPublication
    class Meta:
        unknown=EXCLUDE

class PriorityClaimSchema(BaseSchema):
    __model__ = InpadocPriorityClaim
    doc_type = fields.Str(data_key='@document-id-type')
    kind_code = fields.Str(data_key='@kind')
    date = fields.Date(format="%Y%m%d")
    number = fields.Str(data_key='doc-number')

    class Meta:
        unknown=EXCLUDE

class CpcClassSchema(BaseSchema):
    __model__ = CpcClass
    sequence = fields.Str(data_key='@sequence')
    section = fields.Str()
    cpc_class = fields.Str(data_key='class')
    subclass = fields.Str()
    main_group = fields.Str(data_key='main-group')
    sub_group = fields.Str(data_key='subgroup')
    classification_value = fields.Str(data_key='classification-value')
    generating_office = fields.Str(data_key='generating-office')

    class Meta:
        unknown=EXCLUDE

class UsClassSchema(BaseSchema):
    symbol = fields.Str(data_key = 'classification-symbol')

    class Meta:
        unknown=EXCLUDE

class InpadocBiblioSchema(BaseSchema):
    __model__ = InpadocBiblio

    family_id = fields.Str(data_key='@family-id')
    country = fields.Str(data_key='@country')
    number = fields.Str(data_key='@doc-number')
    kind_code = fields.Str(data_key='@kind')
    abstract = fields.Method(deserialize='get_abstract')
    inventors = fields.Method(deserialize='get_inventors')
    applicants = fields.Method(deserialize='get_applicants')
    publications = ListField(fields.Nested(InpadocPublicationSchema))
    ipc_classes = ListField(fields.Str(), allow_none=True)
    cpc_classes = ListField(fields.Nested(CpcClassSchema), allow_none=True)
    us_classes = ListField(fields.Nested(UsClassSchema), allow_none=True)
    applications = ListField(fields.Nested(InpadocApplicationSchema))
    priority_claims = ListField(fields.Nested(PriorityClaimSchema))
    title = fields.Str()

    def get_abstract(self, data):
        return data['p']

    def get_inventors(self, data):
        return [ i['inventor-name']['name'] for i in data if i['@data-format'] == 'original']

    def get_applicants(self, data):
        return [ i['applicant-name']['name'] for i in data if i['@data-format'] == 'original']

    def pre_load_priority_claims(self, bib):
        pcs = resolve(bib, 'priority-claims.priority-claim')
        out = list()
        pcs = pcs if isinstance(pcs, list) else [pcs,]
        for pc in pcs:
            if isinstance(pc['document-id'], list):
                doc_id = pc['document-id'][0]
            else:
                doc_id = pc['document-id']
            pc = {**pc, **doc_id}
            del pc['document-id']
            out.append(pc)
        return out

    @pre_load
    def pre_load(self, data, *args, **kwargs):
        data = data['exchange-document']
        bib = data['bibliographic-data']
        del data['bibliographic-data']
        data['publications'] = resolve(bib, 'publication-reference.document-id')
        ipc_class = resolve(bib, 'classifications-ipcr.classification-ipcr')
        if ipc_class:
            data['ipc_classes'] = [c['text'] for c in ipc_class]
        classifications = resolve(bib, 'patent-classifications.patent-classification')
        if classifications: 
            data['cpc_classes'] = [c for c in classifications if resolve(c, 'classification-scheme.@scheme') == 'CPCI']
            data['us_classes'] = [c for c in classifications if resolve(c, 'classification-scheme.@scheme') == 'UC']
        data['applications'] = resolve(bib, 'application-reference.document-id')
        data['priority_claims'] = self.pre_load_priority_claims(bib)
        data['applicants'] = resolve(bib, 'parties.applicants.applicant')
        data['inventors'] = resolve(bib, 'parties.inventors.inventor')
        titles = resolve(bib, 'invention-title')
        if isinstance(titles, list):
            data['title'] = next(t['#text'] for t in titles if t['@lang'] == 'en') 
        else:
            data['title'] = titles['#text']
        return data

    class Meta:
        unknown=EXCLUDE


import math
import re
import datetime
import xml.etree.ElementTree as ET
from dateutil.parser import parse as parse_date
from inflection import underscore

class AssignmentParser:
    def parse(self, text):
        tree = ET.fromstring(text.encode("utf-8"))
        result = tree.find("./result")
        num_found = int(result.attrib["numFound"])
        return num_found, [self.doc(doc) for doc in result]

    def doc(self, element):
        data = self.xml_to_dict(element)
        data[
            "image_url"
        ] = f'http://legacy-assignments.uspto.gov/assignments/assignment-pat-{data["display_id"]}.pdf'
        return data

    def xml_to_list(self, element):
        output = list()
        for el in list(element):
            output.append(self.xml_to_pytype(el))
        if len(output) == 1:
            output = output[0]
        return output

    def xml_to_pytype(self, element):
        if element.text == "NULL":
            return None
        elif element.tag == "str":
            return element.text
        elif element.tag == "date":
            # To ensure the result is JSON-serializable, convert to isoformat
            date = parse_date(element.text).date()
            if date < parse_date("1776-01-01").date():
                return None
            return date.isoformat()
        elif element.tag == "int" or element.tag == "long":
            return int(element.text)

    def xml_to_dict(self, element):
        output = dict()
        for el in list(element):
            key = el.attrib["name"]
            key = underscore(key)
            if el.tag == "arr" or el.tag == "lst":
                value = self.xml_to_list(el)
            else:
                value = self.xml_to_pytype(el)
            output[key] = value
        return output


import pytest
import datetime

from patent_client import Inpadoc, Assignment, USApplication
from patent_client.util.deprecated import Manager
Manager.test_mode = True

class TestReadmeExamples():

    def test_us_applications(self):
        app = USApplication.objects.get('15710770')
        assert app.patent_title == 'Camera Assembly with Concave-Shaped Front Face'

    def test_assignments(self):
        assignments = Assignment.objects.filter(assignee='Google')
        assert len(assignments) >= 23860
        assignment = Assignment.objects.get('47086-788')
        assert assignment.conveyance_text == 'ASSIGNMENT OF ASSIGNORS INTEREST'

    def test_inpadoc(self):
        pub = Inpadoc.objects.get('EP3082535A1')
        assert pub.biblio.title == 'AUTOMATIC FLUID DISPENSER'
        assert pub.biblio.priority_claims.values_list('number', flat=True).to_list() == ['US201314137130', 'WO2014US71849']

    @pytest.mark.skip('re-implementing EPO')
    def test_epo(self):
        epo = Epo.objects.get('EP3082535A1')
        assert epo.title == 'AUTOMATIC FLUID DISPENSER'
        assert epo.status ==[{'description': 'Examination is in progress', 'code': '14', 'date': datetime.date(2018, 6, 15)}]


from dataclasses import dataclass, field, asdict, fields
from collections import OrderedDict
from typing import List, Dict, Optional
import datetime
from dateutil.relativedelta import relativedelta
from patent_client.util import one_to_many, one_to_one, Model, QuerySet, ListManager

@dataclass(frozen=True)
class Relationship(Model):
    parent_appl_id: str
    child_appl_id: str
    relationship: str
    child_app_filing_date: Optional[datetime.date] = None
    parent_app_filing_date: Optional[datetime.date] = None
    parent = one_to_one('patent_client.uspto.peds.USApplication', appl_id='parent_appl_id')
    child = one_to_one('patent_client.uspto.peds.USApplication', appl_id='child_appl_id')

@dataclass
class PtaPteHistory(Model):
    number: float
    date: datetime.date
    description: str
    pto_days: int
    applicant_days: int
    start: float

@dataclass
class PtaPteSummary(Model):
    a_delay: int
    b_delay: int
    c_delay: int
    overlap_delay: int
    pto_delay: int
    applicant_delay: int
    pto_adjustments: int
    total_days: int
    kind: Optional[str] = None

@dataclass
class Transaction(Model):
    date: datetime.date
    code: str
    description: str

@dataclass
class Correspondent(Model):
    name: str
    street: Optional[str] = None
    city: Optional[str] = None
    geo_region_code: Optional[str] = None
    postal_code: Optional[str] = None
    country: Optional[str] = None
    cust_no: Optional[str] = None

@dataclass
class Attorney(Model):
    full_name: str
    phone_num: str
    reg_status: Optional[str] = None
    registration_no: Optional[int] = None

@dataclass
class Applicant(Model):
    name: Optional[str] = None
    cust_no: Optional[str] = None
    street: Optional[str] = None
    city: Optional[str] = None
    geo_region_code: Optional[str] = None
    postal_code: Optional[str] = None
    country: Optional[str] = None
    rank_no: Optional[int] = None

@dataclass
class Inventor(Model):
    name: Optional[str] = None
    street: Optional[str] = None
    city: Optional[str] = None
    geo_code: Optional[str] = None
    postal_code: Optional[str] = None
    country: Optional[str] = None
    rank_no: Optional[int] = None
    

@dataclass
class Expiration(Model):
    parent_appl_id: str
    parent_app_filing_date: datetime.date
    parent_relationship: str
    initial_term: datetime.date
    pta_or_pte: float
    extended_term: datetime.date
    terminal_disclaimer_filed: bool

@dataclass
class USApplication(Model):
    __manager__ = 'patent_client.uspto.peds.manager.USApplicationManager'
    appl_id: str = field(compare=True)
    patent_title: Optional[str] = None 
    app_status: Optional[str] = field(default=None, repr=True) 
    inventors: Optional[List[str]] = field(default=None, repr=False) 
    app_filing_date: Optional[datetime.date] = field(default=None, repr=False) 
    app_location: Optional[str] = field(default=None, repr=False) 
    first_inventor_file: Optional[str] = field(default=None, repr=False) 
    app_type: Optional[str] = field(default=None, repr=False) 
    app_entity_status: Optional[str] = field(default=None, repr=False) 
    app_confr_number: Optional[str] = field(default=None, repr=False) 
    applicants: List[Applicant] = field(default_factory=list, repr=False)
    app_status_date: Optional[datetime.date] = field(default=None, repr=False) 
    app_cls_sub_cls: Optional[str] = field(default=None, repr=False) 
    app_grp_art_number: Optional[str] = field(default=None, repr=False) 
    corr_addr_cust_no: Optional[str] = field(default=None, repr=False) 
    app_cust_number: Optional[str] = field(default=None, repr=False) 
    app_attr_dock_number: Optional[str] = field(default=None, repr=False) 
    patent_number: Optional[str] = field(default=None, repr=False) 
    patent_issue_date: Optional[datetime.date] = field(default=None, repr=False) 
    app_early_pub_number: Optional[str] = field(default=None, repr=False) 
    app_early_pub_date: Optional[datetime.date] = field(default=None, repr=False) 
    app_exam_name: Optional[str] = field(default=None, repr=False) 
    wipo_early_pub_number: Optional[str] = field(default=None, repr=False) 
    wipo_early_pub_date: Optional[datetime.date] = field(default=None, repr=False) 

    transactions: List[Transaction] = field(default_factory=list, repr=False)
    child_continuity: ListManager[Relationship] = field(default_factory=ListManager.empty, repr=False)
    parent_continuity: ListManager[Relationship] = field(default_factory=ListManager.empty, repr=False)
    pta_pte_tran_history: List[PtaPteHistory] = field(default_factory=list, repr=False)
    pta_pte_summary: Optional[PtaPteSummary] = field(default=None, repr=False)
    correspondent: Optional[Correspondent] = field(default=None, repr=False)
    attorneys: List[Attorney] = field(default_factory=list, repr=False)

    assignments = one_to_many('patent_client.uspto.assignment.Assignment', appl_id='appl_id')
    trials = one_to_many('patent_client.uspto.ptab.PtabProceeding', appl_id='appl_id')

    @property
    def family(self) -> QuerySet:
        return QuerySet([
            self.child_continuity.values_list('child', flat=True),
            [self, ],
            self.parent_continuity.values_list('parent', flat=True),
        ])

    def __hash__(self):
        return hash(self.appl_id)

    @property
    def kind(self) -> str:
        if "PCT" in self.appl_id:
            return "PCT"
        if self.appl_id[0] == "6":
            return "Provisional"
        return "Nonprovisional"

    @property
    def priority_date(self) -> datetime.date:
        if not self.parent_continuity:
            return self.app_filing_date
        else:
            return sorted(p.parent_app_filing_date for p in self.parent_continuity)[0]

    @property
    def expiration(self) -> Optional[Expiration]:
        if "PCT" in self.appl_id:
            raise NotImplementedError("Expiration date not supported for PCT Applications")
        if not self.patent_number:
            return None
        expiration_data = dict()
        term_parents = [
            p
            for p in self.parent_continuity
            if p.relationship
            not in ["Claims Priority from Provisional Application", "is a Reissue of"]
        ]
        if term_parents:
            term_parent = sorted(term_parents, key=lambda x: x.parent_app_filing_date)[
                0
            ]
            relationship = term_parent.relationship
            parent_filing_date = term_parent.parent_app_filing_date
            parent_appl_id = term_parent.parent_appl_id
        else:
            relationship = "self"
            parent_appl_id = self.appl_id
            parent_filing_date = self.app_filing_date

        expiration_data["parent_appl_id"] = parent_appl_id
        expiration_data["parent_app_filing_date"] = parent_filing_date
        expiration_data["parent_relationship"] = relationship
        expiration_data["initial_term"] = parent_filing_date + relativedelta(years=20) #type: ignore
        expiration_data["pta_or_pte"] = self.pta_pte_summary.total_days if self.pta_pte_summary else 0# type: ignore
        expiration_data["extended_term"] = expiration_data[
            "initial_term"
        ] + relativedelta(days=expiration_data["pta_or_pte"]) # type: ignore

        transactions = self.transactions
        try:
            disclaimer = next(t for t in transactions if t.code == "DIST")
            expiration_data["terminal_disclaimer_filed"] = True
        except StopIteration:
            expiration_data["terminal_disclaimer_filed"] = False

        return Expiration(**expiration_data) # type: ignore




import math
from typing import Iterable, Generic
import inflection

from patent_client.util import Manager, ModelType
from patent_client import session

from .schema import PtabProceedingSchema, PtabDocumentSchema, PtabDecisionSchema
from .model import PtabProceeding, PtabDocument, PtabDecision
from .util import conversions, peds_to_ptab

class PtabManager(Manager, Generic[ModelType]):
    page_size = 25
    instance_schema = None

    def _get_results(self):
        total = self._len()
        offset = self.config["offset"]
        limit = self.config["limit"]
        if limit:
            max_item = total if total - offset < limit else offset + limit
        else:
            max_item = total
        item_range = (offset, max_item)
        page_range = (
            int(offset / self.page_size),
            math.ceil(max_item / self.page_size),
        )
        counter = page_range[0] * self.page_size

        for p in range(*page_range):
            for item in self.get_page(p):
                if item_range[0] <= counter < item_range[1]:
                    yield self.__schema__.load(item)
                counter += 1
                if counter >= max_item:
                    return StopIteration

    def get_page(self, page_no):
        query = self.query()
        query["recordStartNumber"] = page_no * self.page_size
        response = session.get(self.query_url, params=query)
        return response.json()["results"]

    def __len__(self):
        length = self._len() - self.config["offset"]
        if self.config["limit"]:
            return length if length < self.config["limit"] else self.config["limit"]
        else:
            return length

    def _len(self):
        response = session.get(self.query_url, params=self.query())
        return response.json()["recordTotalQuantity"]

    def query(self):
        query = dict()
        for k, v in self.config["filter"].items():
            key = k if k not in peds_to_ptab else peds_to_ptab[k]
            key = inflection.camelize(key, uppercase_first_letter=False)
            query[key] = " ".join(v)
        query["recordTotalQuantity"] = self.page_size
        query["sortOrderCategory"] = " ".join(
            inflection.camelize(o, uppercase_first_letter=False)
            for o in self.config["order_by"]
        )
        return query

class PtabProceedingManager(PtabManager[PtabProceeding]):
    query_url = "https://developer.uspto.gov/ptab-api/proceedings"
    primary_key = "proceeding_number"
    __schema__ = PtabProceedingSchema()


class PtabDocumentManager(PtabManager[PtabDocument]):
    query_url = "https://developer.uspto.gov/ptab-api/documents"
    primary_key = "document_identifier"
    __schema__ = PtabDocumentSchema()

class PtabDecisionManager(PtabManager[PtabDecision]):
    query_url = "https://developer.uspto.gov/ptab-api/decisions"
    primary_key = "identifier"
    __schema__ = PtabDecisionSchema()


# flake8: noqa
__version__ = "2.0.0"

import json
import os
import sys
import platform
import shutil
import time
from pathlib import Path

import requests_cache
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
import datetime


# Set Up Requests Cache
cache_max_age = datetime.timedelta(days=3)
cache_dir = "~/.patent_client"
CACHE_BASE = Path(cache_dir).expanduser()
CACHE_BASE.mkdir(exist_ok=True)
CACHE_CONFIG = dict(
    expire_after=cache_max_age,
    backend=requests_cache.backends.sqlite.DbCache(
        location=str(CACHE_BASE / "requests_cache")
    ),
    allowable_methods=("GET", "POST"),
    )

session = requests_cache.CachedSession(**CACHE_CONFIG)
session.cache.remove_old_entries(datetime.datetime.utcnow() - cache_max_age)
session.headers["User-Agent"] = f"Python Patent Clientbot/{__version__} (pypatent2018@gmail.com)"

# Install a default retry on the session using urrlib3
retry = Retry(total=5, backoff_factor=0.2)
session.mount('https://', HTTPAdapter(max_retries=retry))
session.mount('http://', HTTPAdapter(max_retries=retry))

SETTINGS_FILE = Path("~/.iprc").expanduser()
if not SETTINGS_FILE.exists():
    DEFAULT_SETTINGS = Path(__file__).parent / "default_settings.json"
    shutil.copy(str(DEFAULT_SETTINGS), SETTINGS_FILE)

SETTINGS = json.load(open(SETTINGS_FILE))

from patent_client.epo.inpadoc.model import Inpadoc  # isort:skip
from patent_client.usitc.model import ITCInvestigation, ITCDocument, ITCAttachment
from patent_client.uspto.ptab.model import PtabProceeding, PtabDocument, PtabDecision
from patent_client.uspto.assignment.model import Assignment
from patent_client.uspto.peds.model import USApplication



from dataclasses import dataclass
from typing import List, Optional
import datetime
from collections import OrderedDict

from patent_client import session
from patent_client.util import one_to_one, one_to_many, Model, QuerySet

@dataclass
class Property(Model):
    invention_title: str
    inventors: str
    # Numbers
    appl_id: str
    pct_num: str
    intl_reg_num: str
    publ_num: str
    pat_num: str
    # Dates
    filing_date: datetime.date
    intl_publ_date: datetime.date
    issue_date: datetime.date
    publ_date: datetime.date

    us_application = one_to_one("patent_client.USApplication", appl_id="appl_id")

@dataclass
class Person(Model):
    name: str
    address: str
    city: str
    state: str
    post_code: str
    country_name: str

@dataclass
class Assignor(Model):
    name: str
    ex_date: datetime.date
    date_ack: datetime.datetime

@dataclass
class Assignee(Model):
    name: str
    address: str
    city: str
    state: str
    country_name: str
    postcode: str

@dataclass
class Assignment(Model):
    """
    Assignments
    ===========
    This object wraps the USPTO Assignment API (https://assignments.uspto.gov)

    ----------------------
    To Fetch an Assignment
    ----------------------
    The main way to create an Assignment is by querying the Assignment manager at Assignment.objects

    Assignment.objects.filter(query) -> obtains multiple matching applications
    Assignment.objects.get(query) -> obtains a single matching application, errors if more than one is retreived

    The query can either be a single number, which is treated as a reel/frame number (e.g. "123-1321"), or a keyword.
    Available query types are: 
        patent_number, 
        appl_id (application #), 
        app_early_pub_number (publication #), 
        assignee,
        assignor,
        pct_number (PCT application #),
        correspondent,
        reel_frame

    --------------
    Using the Data
    --------------
    An Assignment object has the following properties:
        id (reel/frame #)
        attorney_dock_num
        conveyance_text
        last_update_date
        page_count
        recorded_date
        correspondent
        assignees
        assignors
        properties

    Additionally, the original assignment document can be downloaded to the working directory by calling:

    assignment.download()
    
    ------------
    Related Data
    ------------
    An Assignment is also linked to other resources available through patent_client. 
    A list of all assigned applications is available at:

    assignment.us_applications

    Additionally, each property entry in properties links to the corresponding application at:

    assignment.properties[0].us_application


    """
    __manager__ = "patent_client.uspto.assignment.manager.AssignmentManager"
    id: str
    conveyance_text: str
    last_update_date: str
    page_count: int
    recorded_date: datetime.date
    corr_name: str
    corr_address: str
    assignors: QuerySet[Assignor]
    assignees: QuerySet[Assignee]
    properties: QuerySet[Property]
    assignment_record_has_images: bool
    transaction_date: Optional[datetime.date] = None
    date_produced: Optional[datetime.date] = None

    us_applications = one_to_many("patent_client.USApplication", appl_id="appl_num")

    @property
    def image_url(self):
        reel, frame = self.id.split('-')
        reel = reel.rjust(6, '0')
        frame = frame.rjust(4, '0')
        return f'http://legacy-assignments.uspto.gov/assignments/assignment-pat-{reel}-{frame}.pdf'

    def download(self):
        response = session.get(self.image_url, stream=True)
        with open(f"{self.id}.pdf", "wb") as f:
            f.write(response.raw.read())






from dataclasses import dataclass
from typing import List, Union, Optional
import datetime
from patent_client.util import Model, one_to_many, one_to_one

@dataclass
class PtabProceeding(Model):
    __manager__ = 'patent_client.uspto.ptab.manager.PtabProceedingManager'
    subproceeding_type_category: str
    proceeding_number: str
    proceeding_status_category: str
    proceeding_type_category: str

    # Party Information
    appl_id: str
    respondent_party_name: str
    petitioner_counsel_name: Optional[str] = None 
    petitioner_party_name: Optional[str] = None
    respondent_counsel_name: Optional[str] = None
    respondent_patent_owner_name: Optional[str] = None

    # Application Information
    inventor: Optional[str] = None
    patent_number: Optional[str] = None
    respondent_technology_center_number: Optional[str] = None
    respondent_grant_date: Optional[datetime.date] = None
    accorded_filing_date: Optional[datetime.date] = None
    decision_date: Optional[datetime.date] = None
    institution_decision_date: Optional[datetime.date] = None
    proceeding_filing_date: Optional[datetime.date] = None

    documents = one_to_many('patent_client.uspto.ptab.PtabDocument', proceeding_number='proceeding_number')
    decisions = one_to_many('patent_client.uspto.ptab.PtabDecision', proceeding_number='proceeding_number')

@dataclass
class PtabDocument(Model):
    __manager__ = 'patent_client.uspto.ptab.manager.PtabDocumentManager'
    document_identifier: str
    document_category: str
    document_type_name: str
    document_number: int
    document_name: str
    document_filing_date: datetime.date
    proceeding_number: str
    proceeding_type_category: str
    title: Optional[str] = None

    proceeding = one_to_one('patent_client.uspto.ptab.PtabProceeding', proceeding_number='proceeding_number')

@dataclass
class PtabDecision(Model):
    __manager__ = 'patent_client.uspto.ptab.manager.PtabDecisionManager' 
    proceeding_number: str
    board_rulings: List[str]
    decision_type_category: str
    document_identifier: str
    document_name: str
    identifier: str
    subdecision_type_category: str
    issue_type: Optional[str] = None
    object_uu_id: Optional[str] = None
    petitioner_technology_center_number: Optional[str] = None
    proceeding = one_to_one('patent_client.uspto.ptab.PtabProceeding', proceeding_number='proceeding_number')





from collections import OrderedDict
from dataclasses import dataclass, fields, make_dataclass, fields
import importlib
import typing

from .manager import QuerySet

ManagerType = typing.TypeVar('ManagerType')

class ModelMeta(type):
    def __new__(cls, name, bases, dct):
        klass = super().__new__(cls, name, bases, dct)
        return klass
    
    @property
    def objects(cls):
        if cls.__manager__ is None:
            return None
        obj_module, obj_class = cls.__manager__.rsplit('.', 1)
        return getattr(importlib.import_module(obj_module), obj_class)()

class ModelABC(object):
    __manager__ = None

@dataclass
class Model(ModelABC, metaclass=ModelMeta):
    def as_dict(self):
        output = OrderedDict()
        for k, v in self:
            if isinstance(v, Model):
                output[k] = v.as_dict()
            elif isinstance(v, (list, QuerySet)):
                output[k] = [i.as_dict() if hasattr(i, 'as_dict') else i for i in v]
            else:
                output[k] = v
        return output

    def fields(self):
        return fields(self)

    def __iter__(self):
        for f in sorted(self.fields(), key=lambda x: x.name):
            if hasattr(self, f.name):
                yield (f.name, getattr(self, f.name))
    
    def to_pandas(self):
        import pandas as pd
        return pd.Series(self.as_dict())



import xml.etree.ElementTree as ET
from urllib import parse
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
import requests_cache

from patent_client import CACHE_CONFIG, CACHE_BASE, SETTINGS

CLIENT_SETTINGS = SETTINGS["ItcEdis"]
USERNAME = CLIENT_SETTINGS["Username"]
PASSWORD = CLIENT_SETTINGS["Password"]
key_file = CACHE_BASE / 'edis_key.txt'

class AuthenticationException(Exception):
    pass

class EdisSession(requests_cache.CachedSession):
    auth_url = "https://edis.usitc.gov/data/secretKey/"

    def __init__(self, *args, username, password, **kwargs):
        super(EdisSession, self).__init__(*args, **kwargs)
        self.username = username
        self.password = password
        self.secret_key = key_file.read_text() if key_file.exists() else ''
        self.auth = (self.username, self.secret_key)
        # Setup Retries
        #retry = Retry(total=1, backoff_factor=0.2)
        #self.mount('https://', HTTPAdapter(max_retries=retry))
        #self.mount('http://', HTTPAdapter(max_retries=retry))

    def request(self, *args, **kwargs):
        response = super(EdisSession, self).request(*args, **kwargs)
        if not response.ok:
            self.authenticate()
            response = super(EdisSession, self).request(*args, **kwargs)
        return response

    def authenticate(self):
        with self.cache_disabled():
            self.auth = None
            response = super(EdisSession, self).request(
                "POST",
                f"{self.auth_url}{self.username}",
                data={"password": self.password}
                )
        if not response.ok:
            raise AuthenticationException(
                "EDIS Authentication Failed! Did you provide the correct username and password?"
            )
        tree = ET.fromstring(response.text)
        self.secret_key = tree.find("secretKey").text
        key_file.write_text(self.secret_key)
        self.auth = (self.username, self.secret_key)

session = EdisSession(username=USERNAME, password=PASSWORD, **CACHE_CONFIG)


import os
from tempfile import TemporaryDirectory
import datetime
import pytest
from .model import Assignment


class TestAssignment:
    def test_fetch_assignments_by_assignee(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assert len(assignments) >= 22
        assert assignments[5].assignors[0].name == "OEHRING, JARED"
        assignments = Assignment.objects.filter(assignee="LogicBlox")
        assert len(assignments) >= 9

    def test_fetch_assignments_by_patent(self):
        assignments = Assignment.objects.filter(patent_number="8,789,601")
        assert len(assignments) >= 1
        assert "48041-605" in [a.id for a in assignments]

    def test_fetch_assignments_by_application(self):
        assignments = Assignment.objects.filter(appl_id="14/190,982")
        assert len(assignments) >= 1

    def test_fetch_assignee_with_greater_than_500_assignments(self):
        assignments = Assignment.objects.filter(assignee="Borealis")
        assert len(assignments) >= 1268

    def test_get_assignment_image(self):
        assignments = Assignment.objects.filter(patent_number=6095661)
        assignment = assignments[0]
        assert (
            assignment.image_url
            == "http://legacy-assignments.uspto.gov/assignments/assignment-pat-038505-0128.pdf"
        )

    def test_slice_assignments(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assignment_list1 = [assignment.id for assignment in assignments[0:5]]
        assert len(assignment_list1) == 5

        assignment_list2 = [assignment.id for assignment in assignments[:5]]
        assert len(assignment_list2) == 5

        assignment_list3 = [assignment.id for assignment in assignments[-5:]]
        assert len(assignment_list3) == 5

    def test_iterate_assignments(self):
        assignments = Assignment.objects.filter(assignee="US Well Services")
        assignment_list = [assignment.id for assignment in assignments]
        assert len(assignment_list) == len(assignments)

    @pytest.mark.skip("Lookup api does not support multiple inputs")
    def test_can_fetch_multiple(self):
        assignments = Assignment.objects.filter(appl_id=['13089872', '15216946'])
        assert assignments.count() == 5


class TestAssignmentBugs:
    def test_id_43433_231(self):
        assignment = Assignment.objects.get("43433-231")
        assert assignment.properties[0].filing_date == datetime.date(2016, 10, 25)

    def test_bug_scidrill(self):
        assignments = Assignment.objects.filter(assignee="Scientific Drilling")
        assignment_list = list(assignments.values_list("appl_num", flat=True))
        assert len(assignment_list) == 58



from __future__ import annotations
import dataclasses as dc
import tempfile
import typing as tp
import datetime as dt
from pathlib import Path
from .session import session

from patent_client.util import Model, one_to_many, one_to_one

@dc.dataclass
class ITCInvestigation(Model):
    __manager__ = 'patent_client.usitc.manager.ITCInvestigationManager'
    number: str
    phase: str
    status: str
    title: str
    type: str
    docket_number: str
    documents = one_to_many("patent_client.ITCDocument", investigation_number="number")
        

@dc.dataclass
class ITCDocument(Model):
    __manager__ = 'patent_client.usitc.manager.ITCDocumentManager'
    id: int
    investigation_number: str
    type: str
    title: str
    security: str
    filing_party: str
    filed_by: str
    filed_on_behalf_of: str
    action_jacket_control_number: str
    memorandum_control_number: str
    date: dt.date
    last_modified: dt.date

    investigation = one_to_one(
        "patent_client.ITCInvestigation", investigation_number="investigation_number"
    )
    attachments = one_to_many("patent_client.ITCAttachment", document_id="id")

@dc.dataclass
class ITCAttachment(Model):
    __manager__ = 'patent_client.usitc.manager.ITCAttachmentManager' 
    id: int
    document_id: int
    title: str
    file_size: int
    file_name: str
    pages: int
    created_date: dt.date
    last_modified_date: dt.date
    document = one_to_one("patent_client.ITCDocument", id="document_id")

    @property
    def download_url(self):
        return f"https://edis.usitc.gov/data/download/{self.document_id}/{self.id}"

    def download(self, file_obj=None):
        *_, ext = self.file_name.split(".")
        if file_obj is None:
            file_obj = tempfile.NamedTemporaryFile(suffix=ext)
        response = session.get(
            self.download_url, stream=True
        )
        with file_obj.open("wb") as f:
            for chunk in response.iter_content(1024):
                f.write(chunk)
        return file_obj


import os
import requests_cache
import datetime as dt
import xmltodict
from pathlib import Path

from patent_client import SETTINGS

cache_max_age = dt.timedelta(days=3)
cache_dir = "~/.patent_client"
CACHE_BASE = Path(cache_dir).expanduser()
CACHE_BASE.mkdir(exist_ok=True)

CACHE_CONFIG = dict(
    expire_after=cache_max_age,
    backend=requests_cache.backends.sqlite.DbCache(
        location=str(CACHE_BASE / "requests_cache")
    ),
    )

CLIENT_SETTINGS = SETTINGS["EpoOpenPatentServices"]
if os.environ.get("EPO_KEY", False):
    KEY = os.environ["EPO_KEY"]
    SECRET = os.environ["EPO_SECRET"]
else:
    KEY = CLIENT_SETTINGS["ApiKey"]
    SECRET = CLIENT_SETTINGS["Secret"]

NS = {
    "http://ops.epo.org": None,
    "http://www.epo.org/exchange": None,
    "http://www.epo.org/fulltext": None,
    "http://www.epo.org/register": None,
}

class OpsSession(requests_cache.CachedSession):
    def __init__(self, *args, key=None, secret=None, **kwargs):
        super(OpsSession, self).__init__(*args, **kwargs)
        self.key: str = key
        self.secret: str = secret
        self.expires: dt.datetime = dt.datetime.utcnow()
    
    def request(self, *args, **kwargs):
        response = super(OpsSession, self).request(*args, **kwargs)
        if response.status_code in (403, 400):
            print('Reauthorizing!')
            auth_response = self.get_token()
            response = super(OpsSession, self).request(*args, **kwargs)
        return response
        
    def get_token(self):
        auth_url = "https://ops.epo.org/3.2/auth/accesstoken"
        response = super(OpsSession, self).request(
            'post',
            auth_url, auth=(self.key, self.secret), data={"grant_type": "client_credentials"}
        )
        data = response.json()
        self.expires = dt.datetime.fromtimestamp(int(data['issued_at']) / 1000) + dt.timedelta(seconds=int(data['expires_in']))
        self.headers['Authorization'] = f"Bearer {data['access_token']}"
        return response

session = OpsSession(
    key=KEY, 
    secret=SECRET,
    **CACHE_CONFIG
)